#############################################################
#                                                           #
#            I, generate data and other general functions   #                            
#                                                           #
#############################################################

## I.1, generate data for scenario I
## n: sample size
## mu1: mean of patients in high risk
## mu0: mean of patients in low risk

genedata_scenario_I     <- function(n, mu1 = 2, mu0 = 1) {
  
  Y                      <- as.matrix(rep(NA, n))
  Y[1:floor(n/2)]        <- 0
  Y[(floor(n/2)+1) : n]  <- 1
  
  X1                      <- rbind(as.matrix(rnorm(floor(n/2), mean = mu0)) , as.matrix(rnorm(n - floor(n/2), mean = mu1) ) )
  X2                      <- rbind(as.matrix(rnorm(floor(n/2), mean = mu0)) , as.matrix(rnorm(n - floor(n/2), mean = mu1) ) )
  X3                      <- as.matrix( rnorm(n, mean = 0, sd = 1) )
  
  
  Marker1                <- cbind(X1, X2, X3)
  
  Marker                 <- cbind(as.matrix(rep(1, n)), X1, X2, X3)
  
  data                   <- data.frame(Y = Y)
  
  data$Marker            <- Marker
  
  data$Marker1           <- Marker1
  
  
  return(data)
}

## I.2, logistic function
  
logistic                 <- function(x){
    
    y                      <- 1/( 1 + exp(-x) ) 
    
    return(y)
    
}

## I.3, generate outliers for scenario II
## n: sample size
## outlier_max is the mean of biomarkers (X1, X2) for low risk patients
## outlier_min is the mean of biomarkers (X1, X2) for high risk patients


genedata_outlier_noise       <- function(n, outlier_max, outlier_min){

Y                      <- as.matrix(rep(NA, n))
Y[1:floor(n/2)]        <- 0
Y[(floor(n/2)+1) : n]  <- 1

X1                      <- rbind(as.matrix(rnorm(floor(n/2), mean = outlier_max)) , as.matrix(rnorm(n - floor(n/2), mean = outlier_min) ))
X2                      <- rbind(as.matrix(rnorm(floor(n/2), mean = outlier_max)) , as.matrix(rnorm(n - floor(n/2), mean = outlier_min)) )
X3                      <- as.matrix( rnorm(n, mean = 0, sd = 1) )

Marker1                <- cbind(X1, X2, X3)

Marker                 <- cbind(as.matrix(rep(1, n)), X1, X2, X3)

data                   <- data.frame(Y = Y)

data$Marker            <- Marker

data$Marker1           <- Marker1

return(data)


}

## I.4, generate data for scenario II
## n: sample size
## prob_outlier: probability of outliers
## outlier_max is the mean of biomarkers (X1, X2) for low risk patients for outliers.
## outlier_min is the mean of biomarkers (X1, X2) for high risk patients for outliers
## mu1_scenario_I: mean of patients in high risk for scenario I.
## mu0_scenario_I: mean of patients in low risk for scenario I.

genedata_scenario_II           <- function(n, prob_outlier, outlier_max, outlier_min, mu1_scenario_I = 2, mu0_scenario_I = 1){
  
  n1                   <- floor( n*prob_outlier )
  n2                   <- n - floor( n*prob_outlier)
  
  data1                <- genedata_outlier_noise(n1, outlier_max, outlier_min)
  data2                <- genedata_scenario_I(n2, mu1_scenario_I, mu0_scenario_I)
  
  
  data                 <- rbind(data1, data2)
  
  
  return(data)
  
}



## I.5, set default values of parameters. 

set_genepara      <- function(mu0_scenario_I          = 1,
                              mu1_scenario_I          = 2,
                              prob_outlier            = 0.05,
                              outlier_max             = 10,
                              outlier_min             = -10
){
  
  gene_para   <- list(mu0_scenario_I          = mu0_scenario_I,
                      mu1_scenario_I          = mu1_scenario_I,
                      prob_outlier            = prob_outlier,
                      outlier_max             = outlier_max,
                      outlier_min             = outlier_min
  )
  
  return(gene_para)
  
}

#############################################################
#                                                           #
#            II, Standard Approach                          #                            
#                                                           #
#############################################################

## II.1, compute TPF - lambda * FPF for linear rule 
## data: the data set
## beta: the parameter for the linear rule.
## lambda: the weight in the difference

caldiff_TPFP_linklinear    <- function(data, beta, lambda) {
  
  f            <- data$Marker %*% beta
  
  data$f_index <- ifelse(f > 0, 1, 0)
  
  data_1       <- data[data$Y==1, ]
  data_0       <- data[data$Y==0, ]
  
  TP           <- mean(data_1$f_index)
  FP           <- mean(data_0$f_index)
  
  Diff         <- TP - lambda * FP
  
  return( list(TP = TP, FP = FP, Diff = Diff) )
  
}

## II.2, the Standard Approach
## data: the data set
## alpha: the constraint for TPF
## KK: the number of folds in cross-validation

calpara_constraint_TPFP_mle   <- function(data, alpha, KK){
  
  n_o           <- dim(data$Marker)[1]
  index         <- sample(KK, n_o, replace = TRUE)
  
  C_temp      <- c()
  
  for (kk in 1:KK) {
    
    training_data          <- data[index != kk, ]
    test_data              <- data[index == kk, ]
    
    
    beta_0 <- as.matrix(glm(training_data$Y ~ training_data$Marker1, family ="quasibinomial")$coefficients)
    
    C      <- sort(test_data$Marker %*% beta_0)
    
    n      <- length(C)
    
    FP   <- rep(NA, n)
    
    for (ii in 1:n) {
      
      c            <- C[ii]
      beta_temp    <- beta_0
      beta_temp[1] <- beta_0[1] - c
      FP[ii]       <- caldiff_TPFP_linklinear(test_data, beta_temp, 1)$FP
      
      
    }
    
    temp              <- FP - alpha
    temp[temp > 0]    <- 1
    
    c_temp           <- min(C[abs( temp ) == min(abs(temp))])
    
    C_temp       <- c(C_temp, c_temp)
    
  }
  
  c      <- mean(C_temp)
  
  
  beta_0 <- as.matrix(glm(data$Y ~ data$Marker1, family ="quasibinomial")$coefficients)
  
  beta        <- beta_0
  beta[1]     <- beta_0[1] - c
  
  beta        <- beta/abs(beta[1])
  
  
  return(beta)
  
}

#############################################################
#                                                           #
#            III, Plug-in Approach                          #                            
#                                                           #
#############################################################

## III.1, calculate the result for plugin method for fixed estimating method
## data: training data set
## K: number of folds in cross-validation
## alpha: constraint for FPF
## fitting_method: choose "randomforest" or "logistic"

calpara_constraint_plugin   <- function(data, K, alpha, fitting_method = "logistic"){
  
  n           <- dim(data)[1]
  index       <- sample(K, n, replace = TRUE)
  prob        <- rep(0, n)
  
  
  for (kk in 1:K) {
    
    training_data <- data[index != kk, ]
    test_data     <- data[index == kk, ]
    
    if (fitting_method == "logistic") {  
      theta         <- (glm(training_data$Y ~ training_data$Marker - 1, family = "quasibinomial") )$coefficients
      
      p_test        <- logistic(test_data$Marker %*% theta)
      
    } else if (fitting_method == "randomforest") {
      
      model     <- randomForest(x      = training_data$Marker,
                                y      = factor(training_data$Y),
                                xtest  = test_data$Marker,
                                ytest  = factor(test_data$Y),
                                keep.forest=TRUE,
                                importance = TRUE)
      
      p_test    <- predict(model, test_data$Marker, "prob")[, 2]
      
      
    }
    
    prob[index == kk] = p_test
    
  }
  
  lambda          <- callambda_constraint_plugin(data, alpha, prob)
  
  
  if (fitting_method == "logistic") {  
    
    theta         <- (glm(data$Y ~ data$Marker - 1, family = "quasibinomial") )$coefficients
    
  } else if (fitting_method == "randomforest") {
    
    theta <- 0
  }
  
  return(list(lambda = lambda, theta = theta))
  
}


## III.2, choose lamba for fixed alpha and p (probability of D =1 | biomarker)
## data: data set
## alpha: constraint for FPF
## p: probability of D =1 | biomarker

callambda_constraint_plugin  <- function(data, alpha, p){
  
  lambda_low   <- 0
  lambda_up    <- 100
  
  
  delta         <- 1
  ii            <- 1
  ALPHA         <- c()
  LAMBDA        <- c()
  
  while( delta >= 10e-6 && ii <= 200){
    
    
    
    lambda_temp  <- (lambda_low + lambda_up)/2
    LAMBDA       <- cbind(LAMBDA, lambda_temp)
    alpha_temp   <- calalpha_updatelambda_constraint_plugin(data, lambda_temp, p)
    ALPHA        <- cbind(ALPHA, alpha_temp)
    
    
    if (alpha_temp > alpha) {
      lambda_low   <- lambda_temp  
      
      
    } else {
      
      lambda_up   <- lambda_temp
      
    }
    
    delta        <- max( abs(lambda_up - lambda_low) )
    
    ii           <- ii + 1  
    
  }
  
  if (sum(ALPHA > alpha) == length(ALPHA)) {
    
    lambda  <- lambda_up
    
  } else {
    
    lambda  <- min(LAMBDA[ALPHA <= alpha])
  }
  
  return(lambda)
  
}




#### III.3, calculate FPF for fixed lambda and p
## data: data set
## alpha: constraint for FPF
## p: probability of D =1 | biomarker


calalpha_updatelambda_constraint_plugin   <- function(data, lambda, p) {
  
  data_1       <- data[data$Y==1, ]
  data_0       <- data[data$Y==0, ]
  
  p1           <- dim(data_1)[1]/dim(data)[1]
  p0           <- 1 - p1
  
  f            <- 1/p1 * p - lambda/p0 *(1 - p)
  
  data$f_index <- ifelse(f > 0, 1, 0)
  
  data_1       <- data[data$Y==1, ]
  data_0       <- data[data$Y==0, ]
  
  
  output        <- mean(data_0$f_index)
  
  return(output)
  
}

#### III.4, compute TPF and FPF for the validation set by plugin method.
## data1: training data set
## data2: validation data set
## lambda: fixed lambda value
## theta: the estimated parameter in the linear rule for `logistic` method
## fitting_method: different fitting methods


calTPFP_constraint_plugin  <- function(data1, data2, lambda, theta, fitting_method = "logistic") {
  
  if (fitting_method == "logistic") {
    
    p          <- logistic(data2$Marker %*% theta)
    
  } else if (fitting_method == "randomforest") {
    
    model     <- randomForest(x      = data1$Marker,
                              y      = factor(data1$Y),
                              xtest  = data2$Marker,
                              ytest  = factor(data2$Y),
                              keep.forest=TRUE,
                              importance = TRUE)
    
    p    <- predict(model, data2$Marker, "prob")[, 2]
  }
  
  p1           <- mean(data1$Y)
  p0           <- 1 - p1
  
  f            <- 1/p1 * p - lambda/p0 *(1 - p)
  
  data2$f_index <- ifelse(f > 0, 1, 0)
  
  data_1       <- data2[data2$Y==1, ]
  data_0       <- data2[data2$Y==0, ]
  
  TP           <- mean(data_1$f_index)
  FP           <- mean(data_0$f_index)
  
  return( list(TP = TP, FP = FP) )
  
}

### III.5, plug-in-cv, choose the estimating method between random forest and logistic regression
## data1: training data set
## data2: validation data set
## DD: number of folds for cross-validation for choosing
## alpha: constraint for FPF
## KK: number of folds for cross-validation for estimating 

## we set the choosing boundary as alpha +/- 1/5*alpha, which can be modified according to different situations.


calTPFP_constraint_plugin_choose  <- function(data1, data2, DD, alpha, KK){
  
  ## start a cross-validation to choose the fitting_method
  n1          <- dim(data1)[1]
  index       <- sample(DD, n1, replace = TRUE)
  TP_rf       <- 0
  TP_lg       <- 0
  FP_rf       <- 0
  FP_lg       <- 0
  
  
  for (kk in 1:DD){
    
    training_data     <- data1[index != kk, ]
    
    test_data          <- data1[index == kk, ]
    
    para_TPFP_constraint_plugin_rf    <- calpara_constraint_plugin(training_data, KK, alpha, fitting_method = "randomforest")
    
    
    TPFP_constraint_plugin_rf         <- calTPFP_constraint_plugin(training_data, test_data, para_TPFP_constraint_plugin_rf$lambda, 
                                                                   para_TPFP_constraint_plugin_rf$theta, fitting_method = "randomforest")
    TP_rf            <- TP_rf + TPFP_constraint_plugin_rf$TP/DD
    
    FP_rf            <- FP_rf + TPFP_constraint_plugin_rf$FP/DD
    
    
    para_TPFP_constraint_plugin_lg    <- calpara_constraint_plugin(training_data, KK, alpha, fitting_method = "logistic")
    
    TPFP_constraint_plugin_lg         <- calTPFP_constraint_plugin(training_data, test_data, para_TPFP_constraint_plugin_lg$lambda, 
                                                                   para_TPFP_constraint_plugin_lg$theta, fitting_method = "logistic")
    TP_lg            <- TP_lg + TPFP_constraint_plugin_lg$TP/DD
    
    FP_lg            <- FP_lg + TPFP_constraint_plugin_lg$FP/DD
    
    
  }
  
  ## use the data2 to get the Diff
  
  if ( abs(FP_lg - alpha) <= alpha/5 && abs(FP_rf - alpha) <= alpha/5 ){
    
    
    if (TP_lg >= TP_rf)  {
      
      meth = "lg"
      
    } else {
      
      
      meth = "rf"
      
    }
    
  } else if ( abs(FP_lg - alpha) <= alpha/5 && abs(FP_rf - alpha) > alpha/5) {
    
    
    meth = "lg"
    
  } else if ( abs(FP_rf - alpha) <= alpha/5 && abs(FP_lg - alpha) > alpha/5) {
    
    
    meth = "rf"
    
  } else {
    
    
    
    if (TP_lg >= TP_rf)  {
      
      meth = "lg"
      
    } else {
      
      
      meth = "rf"
      
    }
    
  }
  
  return( meth )
  
}



#############################################################
#                                                           #
#            IV. Direct-Optimization Approach               #                            
#                                                           #
#############################################################


## IV.1.a, logistic loss

cal_loss_logistic <- function(x) {
  
  y <- as.matrix( log(1 + exp(-x)))
  
  return(y)
}

## IV.1.b, first derivative, logistic loss

cal_first_deriv_loss_logistic <- function(x) {
  
  y <- as.matrix( logistic(x) -1)
  
  return(y)
  
}


## IV.1.c, second derivative, logistic loss

cal_second_deriv_loss_logistic <- function(x) {
  
  p  <- logistic(x)
  
  y  <- as.matrix( p*(1-p) )
  
  return(y)
  
}

## IV.1.d, hinge loss

cal_loss_hinge <- function(x) {
  
  y  <- as.matrix( c( 1:length(x) ) )
  
  for (ii in 1:length(x) )   {
    
    if (x[ii] <= 1){
      
      y[ii]  <- (1 - x[ii])
      
    } else {
      
      y[ii]  <- 0
      
    } 
    
  }
  
  return (y)
  
}

## IV.1.e, first derivative, hinge loss

cal_first_deriv_loss_hinge <- function(x) {
  
  y  <- as.matrix( c( 1:length(x) ) )
  
  for (ii in 1:length(x) )   {
    
    if (x[ii] <= 1){
      y[ii]  <- - 1
      
    } else {
      
      y[ii]  <- 0
      
    } 
    
  }
  
  return (y)
  
}

## IV.1.f, second derivative, hinge loss

cal_second_deriv_loss_hinge <- function(x) {
  
  y  <- as.matrix( rep(0, length(x)) )
  
  
  return (y)
  
}

## IV.2, calculate the subgradient value for given beta0, beta1, lambda
## data: data set
## beta0: intercept in the linear rule.
## beta1: slope in the linear rule.
## lambda: fixed parameter in TPF - lambda * FPF
## approx_method: different surrogate functions

calS_beta1_maxdiff_TPFP_Subgradient   <- function(data, beta0, beta1, lambda, approx_method = "hinge"){
  
  beta     <- rbind(as.matrix(beta0), beta1)
  
  
  data$f   <- data$Marker %*% beta
  
  data_1   <- data[data$Y==1, ]
  data_0   <- data[data$Y==0, ]
  
  n_1      <- dim(data_1)[1]
  n_0      <- dim(data_0)[1]
  
  if  (approx_method == "logistic") {
    
    S =  1/n_1 * t(data_1$Marker1)%*% (cal_first_deriv_loss_logistic( data_1$f)) -
      lambda/n_0 * t(data_0$Marker1)%*% (cal_first_deriv_loss_logistic( - data_0$f))
    
    
    
  }  else if (approx_method == "hinge") {
    
    S =  1/n_1 * t(data_1$Marker1)%*% (cal_first_deriv_loss_hinge(  data_1$f)) -
      lambda/n_0 * t(data_0$Marker1)%*% (cal_first_deriv_loss_hinge( - data_0$f))
    
    
    
  }
  
  return( S )
}


## IV.3, calculate a series of beta1 for the subgradient method
## data: data set
## beta0: intercept in the linear rule.
## beta1: slope in the linear rule.
## n_S: how many steps in subgradient method.
## a_S: step size in subgradient method. 
## lambda: fixed parameter in TPF - lambda * FPF
## approx_method: different surrogate functions

calbeta1_maxdiff_TPFP_BETA_Subgradient <- function(data, beta0, beta1, n_S, a_S, lambda, approx_method = "hinge") {
  
  
  beta1_old                  <- beta1
  pp                         <- length(beta1)
  delta                      <- 1
  ii                         <- 1
  BETA                       <- beta1_old
  
  while( delta >= 10e-7 && ii <= n_S ) {
    
    S            <- calS_beta1_maxdiff_TPFP_Subgradient(data, beta0, beta1_old, lambda, approx_method)
    
    if (anyNA(S) > 0 | sum(is.infinite(S)) >0) {
      
      beta_new   <- as.matrix(rep(0, pp))
      ii          <- n_S
      
    } else {
      
      
      beta_new    <- beta1_old - a_S* S
      
      if ( sum(is.infinite(beta_new)) >0  ) {
        
        beta_new   <- as.matrix(rep(0, pp))
        ii          <- n_S 
        
      }
      
    }
    
    BETA         <- cbind(BETA, beta_new)
    
    delta        <- max( abs(beta_new - beta1_old) )
    
    ii           <- ii + 1
    
    beta1_old     <- beta_new
    
  }
  
  
  return(BETA1 = BETA)
  
  
}



### IV.4, calculate the objective function TPF - lambda* FPF for a series beta1
## data: data set
## BETA: matrix contains a series of beta
## lambda: fixed parameter in TPF - lambda * FPF
## approx_method: different surrogate functions

cal_Lagrangian_phi_objective_withoutalpha    <- function(data, BETA, lambda, approx_method = "hinge") {
  
  pp           <- dim(BETA)[2]
  Obj          <- as.matrix(rep(0, pp))
  
  if (length(lambda) == 1) {
    
    lambda     <- rep(lambda, pp)
    
  }
  
  for (ii in 1:pp ) {
    
    beta         <- BETA[, ii]
    data$f       <- data$Marker %*% beta
    
    
    ### difference approlambdamating methods
    if  (approx_method == "logistic") {
      
      data$f_phi_p <- cal_loss_logistic(data$f)
      data$f_phi_n <- cal_loss_logistic(-data$f)
      
      
    }  else if (approx_method == "hinge") {
      
      data$f_phi_p <- cal_loss_hinge(data$f)
      data$f_phi_n <- cal_loss_hinge(-data$f)
      
    } 
    
    
    
    data_1       <- data[data$Y==1, ]
    data_0       <- data[data$Y==0, ]
    
    TP_phi       <- mean(data_1$f_phi_p)
    FP_phi       <- mean(data_0$f_phi_n)
    
    Obj[ii]      <- TP_phi + lambda[ii] * FP_phi
    
  }
  
  
  return( Obj )
  
}



## IV.5, calculate beta1 for given beta_0, n_S and a_S
## data: data set
## beta0: intercept in the linear rule.
## beta1: initial for the slope in the linear rule.
## KK: number of folds in cross-validation.
## n_S: how many steps in subgradient method.
## a_S: step size in subgradient method. 
## lambda: fixed parameter in TPF - lambda * FPF
## approx_method: different surrogate functions


calbeta1_maxdiff_TPFP_Subgradient  <- function(data, beta0, beta1, KK, n_S, a_S, lambda, approx_method = "hinge"){
  
  n           <- dim(data)[1]
  pp          <- dim(data$Marker1)[2]
  
  index       <- sample(KK, n, replace = TRUE)
  BETA        <- matrix(rep(0, KK*pp), ncol = KK)
  
  
  for (kk in 1:KK) {
    
    training_data <- data[index != kk, ]
    test_data     <- data[index == kk, ]
    
    
    BETA1_temp    <- calbeta1_maxdiff_TPFP_BETA_Subgradient(training_data, beta0, beta1, n_S, a_S, lambda, approx_method)
    BETA_temp     <- rbind(t(as.matrix(rep(beta0, dim(BETA1_temp)[2]))), BETA1_temp)
    
    Diff_temp     <- cal_Lagrangian_phi_objective_withoutalpha(test_data, BETA_temp, lambda, approx_method)
    XX            <- c(1:length(Diff_temp))
    
    BETA[, kk]    <- as.matrix(BETA1_temp[, min(XX[Diff_temp == min(Diff_temp)])])
    
  }
  
  beta_final   <- as.matrix( apply(BETA, 1, mean) )
  
  return(beta1 = beta_final)
  
  
}


## IV.6, calculate FPF for the linear link for a series of beta
## data: data set
## BETA: matrix contains a series of beta


calFP_linklinear    <- function(data, BETA) {
  
  pp           <- dim(BETA)[2]
  FP           <- as.matrix(rep(0, pp))
  
  for (ii in 1:pp ) {
    
    beta         <- BETA[, ii]
    f            <- data$Marker %*% beta
    
    data$f_index <- ifelse(f > 0, 1, 0)
    
    data_0       <- data[data$Y==0, ]
    
    FP[ii]       <- mean(data_0$f_index)
    
  }
  return( FP )
  
}


### IV.7, calculate FPF for given beta_0, n_S and a_S, and intial guess of beta1
## data: data set
## beta0: intercept in the linear rule.
## beta1: initial for the slope in the linear rule.
## alpha: constrait for FPF
## KK: number of folds in cross-validation.
## n_S: how many steps in subgradient method.
## a_S: step size in subgradient method. 
## lambda: fixed parameter in TPF - lambda * FPF
## approx_method: different surrogate functions

calFP_lambda_constraint_Subgradient <- function(data, beta0, beta1, alpha, KK, n_S, a_S, lambda, approx_method = "squared_hinge"){
  
  
  beta1_temp <- calbeta1_maxdiff_TPFP_Subgradient(data, beta0, beta1, KK, n_S, a_S, lambda, approx_method)
  
  beta_temp  <- rbind(as.matrix(beta0), beta1_temp)
  
  FP         <- calFP_linklinear(data, beta_temp)
  
  return(FP)
  
}


### IV.8, calculate FPF for a series of Lambda, and given beta_0, n_S and a_S
## data: data set
## beta0: intercept in the linear rule.
## beta1: initial for the slope in the linear rule.
## alpha: constraint for FPF
## KK: number of folds in cross-validation.
## n_S: how many steps in subgradient method.
## a_S: step size in subgradient method. 
## lambda_low: lower value for lambda.
## lambda_up: upper value for lambda.
## grid_number: how many values for lambda (from lambda_low to lambda_up)
## approx_method: different surrogate functions

calpara_Lambda_constraint_Subgradient <- function(data, beta0, beta1, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, approx_method = "hinge"){
  
  if (lambda_low == lambda_up) {
    Lambda = rep(lambda_up, grid_number)
    FP     = rep(calFP_lambda_constraint_Subgradient(data, beta0, beta1, alpha, KK, n_S, a_S, lambda_up, approx_method), grid_number)
    
    
  } else {
    
    Lambda <- seq(lambda_low, lambda_up, by = (lambda_up - lambda_low)/(grid_number -1))
    FP     <- rep(0, grid_number)
    
    for (ii in 1:grid_number) {
      
      FP[ii]    <- calFP_lambda_constraint_Subgradient(data, beta0, beta1, alpha, KK, n_S, a_S, Lambda[ii], approx_method)
      
    }
    
    if (sum(FP >= alpha) > 0) {
      Lambda_low <- Lambda[FP >= alpha]
      
      lambda_low <- max(Lambda_low)
    }
    
    
    if (sum(FP < alpha) > 0) {
      Lambda_up <- Lambda[FP < alpha]
      
      lambda_up <- min(Lambda_up)
    }
  }
  
  
  return(list(Lambda = Lambda, FP = FP, lambda_low = lambda_low, lambda_up = lambda_up))
  
  
}



## IV.9, calculate beta1 for given beta_0, n_S and a_S
## data: data set
## beta0: intercept in the linear rule.
## beta1: initial for the slope in the linear rule.
## alpha: constraint for FPF
## KK: number of folds in cross-validation.
## n_S: how many steps in subgradient method.
## a_S: step size in subgradient method. 
## lambda_low: lower value for lambda.
## lambda_up: upper value for lambda.
## grid_number: how many values for lambda (from lambda_low to lambda_up)
## S_lambda: limit of loop
## approx_method: different surrogate functions

calbeta1_constraint_Subgradient <- function(data, beta0, beta1, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method = "hinge"){
  
  ## make sure range of FP by lambda_low and lambda_up contains alpha
  
  
  while ( calFP_lambda_constraint_Subgradient(data, beta0, beta1, alpha, KK, n_S, a_S, lambda_up, approx_method) > alpha && lambda_up <= 50) {
    
    
    lambda_up   <- lambda_up * 2
    
    
  }
  
  
  while ( calFP_lambda_constraint_Subgradient(data, beta0, beta1, alpha, KK, n_S, a_S, lambda_low, approx_method) < alpha && lambda_low >= 10e-06 ) {
    
    lambda_low   <- lambda_low/10
    
  }
  
  
  ### start the loop
  for (jj in 1:S_lambda) {
    
    temp <- calpara_Lambda_constraint_Subgradient(data, beta0, beta1, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, approx_method)
    
    if ( abs(temp$lambda_up - temp$lambda_low) <= 0.01) {
      
      jj     <- S_lambda
      
    }
    
    jj <- jj + 1
    
    lambda_low <- temp$lambda_low
    lambda_up  <- temp$lambda_up
    
    
  }
  
  
  # lambda <- (temp$lambda_up + temp$lambda_low)/2
  
  lambda <- temp$lambda_up
  
  beta1   <- calbeta1_maxdiff_TPFP_Subgradient(data, beta0, beta1, KK, n_S, a_S, lambda, approx_method)
  
  
  return( list( beta1 = beta1, lambda = lambda))
  
  
}



### IV.10,calculate the objective function for TPF - lambda*(FPF - alpha)
## data: data set
## BETa: a series of beta
## lambda: weight
## alpha: constraint for FPF

cal_Lagrangian_objective    <- function(data, BETA, lambda, alpha) {
  
  pp           <- dim(BETA)[2]
  Obj          <- as.matrix(rep(0, pp))
  
  if (length(lambda) == 1) {
    
    lambda     <- rep(lambda, pp)
    
  }
  
  for (ii in 1:pp ) {
    
    beta         <- BETA[, ii]
    f            <- data$Marker %*% beta
    
    data$f_index <- ifelse(f > 0, 1, 0)
    
    data_1       <- data[data$Y==1, ]
    data_0       <- data[data$Y==0, ]
    
    TP           <- mean(data_1$f_index)
    FP           <- mean(data_0$f_index)
    
    Obj[ii]      <- TP - lambda[ii] * (FP- alpha)
    
  }
  
  
  return( Obj )
  
}


## IV.11, calculate beta1 for changing initial guess(some of them are generated randomly)
## data: data set
## beta1_0: initial intercept in the linear rule.
## beta1_1: initial for the slope in the linear rule.
## alpha: constraint for FPF
## ratio: used in choosing the initial guess
## KK: number of folds in cross-validation.
## n_S: how many steps in subgradient method.
## a_S: step size in subgradient method. 
## lambda_low: lower value for lambda.
## lambda_up: upper value for lambda.
## grid_number: how many values for lambda (from lambda_low to lambda_up)
## S_lambda: limit of loop in finding lambda
## KK: number of folds in cross-validation
## K_L: number of folds in cross-validation
## beta_index: whether we input a known initial guess for beta1, '1' means yes, '0' means no.
## beta_number: how many addtional initial guess of beta1 being generated randomly
## approx_method: different surrogate functions


calpara_constraint_Subgradient_varystartpoint   <- function(data, beta1_0, beta1_1, alpha, ratio = 2, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, KK = 5, K_L = 5, beta_index = 1, beta_number = 1, approx_method = "hinge"){
  
  
  n           <- dim(data$Marker)[1]
  pp          <- dim(data$Marker1)[2]
  index       <- sample(K_L, n, replace = TRUE)
  
  
  if (beta_index == 0) {
    
    DD   <- beta_number 
    
  } else {
    
    DD  <- beta_number + 2
    
  }
  
  TP_n1  <- rep(0, DD)
  
  TP_p1  <- rep(0, DD)
  
  FP_n1  <- rep(0, DD)
  
  FP_p1  <- rep(0, DD)
  
  Obj_n1  <- rep(0, DD)
  
  Obj_p1  <- rep(0, DD)
  
  
  BETA    <- matrix(rep(0, DD*pp), nrow = pp)
  
  if (beta_index == 0){
    
    for (ii in 1:DD) {
      
      beta_org    <- as.matrix(rnorm(pp))
      
      BETA[, ii]  <- beta_org
      
      for (kk in 1:K_L) {
        
        training_data <- data[index != kk, ]
        test_data     <- data[index == kk, ]
        
        temp_p1          <- calbeta1_constraint_Subgradient(training_data, 1, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
        
        TP_p1[ii]        <- TP_p1[ii] +  caldiff_TPFP_linklinear(test_data, rbind(as.matrix(1), temp_p1$beta1), 1)$TP/K_L
        
        FP_p1[ii]        <- FP_p1[ii] +  caldiff_TPFP_linklinear(test_data, rbind(as.matrix(1), temp_p1$beta1), 1)$FP/K_L
        
        Obj_p1[ii]       <- Obj_p1[ii] + cal_Lagrangian_objective(test_data, rbind(as.matrix(1), temp_p1$beta1), temp_p1$lambda, alpha) 
        
        
        
        temp_n1          <- calbeta1_constraint_Subgradient(training_data, -1, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
        
        TP_n1[ii]        <- TP_n1[ii] + caldiff_TPFP_linklinear(test_data, rbind(as.matrix(-1), temp_n1$beta1), 1)$TP/K_L
        
        FP_n1[ii]        <- FP_n1[ii] + caldiff_TPFP_linklinear(test_data, rbind(as.matrix(-1), temp_n1$beta1), 1)$FP/K_L
        
        Obj_n1[ii]       <- Obj_n1[ii] + cal_Lagrangian_objective(test_data, rbind(as.matrix(-1), temp_n1$beta1), temp_n1$lambda, alpha) 
        
        
      }
    }
    
  } else {
    
    
    BETA[, 1]  <- beta1_0
    
    BETA[, 2]  <- beta1_1
    
    
    
    for (kk in 1:K_L) {
      
      training_data <- data[index != kk, ]
      test_data     <- data[index == kk, ]
      
      beta_org      <- beta1_0
      
      temp_p1       <- calbeta1_constraint_Subgradient(training_data, 1, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
      
      TP_p1[1]      <- TP_p1[1] +  caldiff_TPFP_linklinear(test_data, rbind(as.matrix(1), temp_p1$beta1), 1)$TP/K_L
      
      FP_p1[1]      <- FP_p1[1] +  caldiff_TPFP_linklinear(test_data, rbind(as.matrix(1), temp_p1$beta1), 1)$FP/K_L
      
      Obj_p1[1]    <- Obj_p1[1] + cal_Lagrangian_objective(test_data, rbind(as.matrix(1), temp_p1$beta1), temp_p1$lambda, alpha) 
      
      
      
      
      
      temp_n1       <- calbeta1_constraint_Subgradient(training_data, -1, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
      
      TP_n1[1]      <- TP_n1[1] + caldiff_TPFP_linklinear(test_data, rbind(as.matrix(-1), temp_n1$beta1), 1)$TP/K_L
      
      FP_n1[1]      <- FP_n1[1] + caldiff_TPFP_linklinear(test_data, rbind(as.matrix(-1), temp_n1$beta1), 1)$FP/K_L
      
      Obj_n1[1]     <- Obj_n1[1] + cal_Lagrangian_objective(test_data, rbind(as.matrix(-1), temp_n1$beta1), temp_n1$lambda, alpha) 
      
      
      
      beta_org      <- beta1_1
      
      temp_p1       <- calbeta1_constraint_Subgradient(training_data, 1, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
      
      TP_p1[2]      <- TP_p1[2] +  caldiff_TPFP_linklinear(test_data, rbind(as.matrix(1), temp_p1$beta1), 1)$TP/K_L
      
      FP_p1[2]      <- FP_p1[2] +  caldiff_TPFP_linklinear(test_data, rbind(as.matrix(1), temp_p1$beta1), 1)$FP/K_L
      
      Obj_p1[2]     <- Obj_p1[2] + cal_Lagrangian_objective(test_data, rbind(as.matrix(1), temp_p1$beta1), temp_p1$lambda, alpha) 
      
      
      
      
      
      temp_n1       <- calbeta1_constraint_Subgradient(training_data, -1, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
      
      TP_n1[2]      <- TP_n1[2] + caldiff_TPFP_linklinear(test_data, rbind(as.matrix(-1), temp_n1$beta1), 1)$TP/K_L
      
      FP_n1[2]      <- FP_n1[2] + caldiff_TPFP_linklinear(test_data, rbind(as.matrix(-1), temp_n1$beta1), 1)$FP/K_L
      
      Obj_n1[2]     <- Obj_n1[2] + cal_Lagrangian_objective(test_data, rbind(as.matrix(-1), temp_n1$beta1), temp_n1$lambda, alpha) 
      
      
      
      
    }
    
    for (ii in 3:DD) {
      beta_org  <- as.matrix(rnorm(pp))
      
      BETA[, ii] <- beta_org
      
      
      for (kk in 1:K_L) {
        
        training_data <- data[index != kk, ]
        test_data     <- data[index == kk, ]
        
        temp_p1          <- calbeta1_constraint_Subgradient(training_data, 1, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
        
        TP_p1[ii]        <- TP_p1[ii] +  caldiff_TPFP_linklinear(test_data, rbind(as.matrix(1), temp_p1$beta1), 1)$TP/K_L
        
        FP_p1[ii]        <- FP_p1[ii] +  caldiff_TPFP_linklinear(test_data, rbind(as.matrix(1), temp_p1$beta1), 1)$FP/K_L
        
        Obj_p1[ii]       <- Obj_p1[ii] + cal_Lagrangian_objective(test_data, rbind(as.matrix(1), temp_p1$beta1), temp_p1$lambda, alpha) 
        
        
        
        
        
        temp_n1          <- calbeta1_constraint_Subgradient(training_data, -1, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
        
        TP_n1[ii]        <- TP_n1[ii] + caldiff_TPFP_linklinear(test_data, rbind(as.matrix(-1), temp_n1$beta1), 1)$TP/K_L
        
        FP_n1[ii]        <- FP_n1[ii] + caldiff_TPFP_linklinear(test_data, rbind(as.matrix(-1), temp_n1$beta1), 1)$FP/K_L
        
        Obj_n1[ii]       <- Obj_n1[ii] + cal_Lagrangian_objective(test_data, rbind(as.matrix(-1), temp_n1$beta1), temp_n1$lambda, alpha) 
        
        
        
        
        
        
      }
      
    }
    
    
  }
  
  
  xx      <- c(1:(2*DD))
  yy      <- rep(0, 2*DD)
  
  BETA_F  <- rbind(t(as.matrix(c(rep(-1, DD), rep(1, DD)) )), cbind(BETA, BETA))
  
  
  TP      <- c(TP_n1, TP_p1)
  FP      <- c(FP_n1, FP_p1)
  Obj     <- c(Obj_n1, Obj_p1)
  
  for (ii in 1:(2*DD)){
    
    if (FP[ii] <= ratio*alpha && FP[ii] >= alpha/ratio){
      
      yy[ii] <- 1
      
    }
    
  }
  
  
  R   <- rep(0, 2*DD)
  
  for (ii in 1:(2*DD)){
    
    if (FP[ii] > alpha){
      
      R[ii] <- FP[ii]/alpha
      
    } else {
      
      if (FP[ii] <= 0.0001) {
        
        R[ii] <- 1000
        
      } else {
        
        R[ii] <- alpha/FP[ii]
        
      }
    }
    
    
  }
  
  
  if (sum(yy) >= 1){
    
    zz       <- xx[yy == 1]
    TP[-zz]  <- 0
    
    dd       <- min(xx[TP == max(TP)])
    
    beta_org <- as.matrix(BETA_F[-1, dd])
    
    beta0    <- BETA_F[1, dd]
    
    
  } else {
    
    
    dd <- min(xx[R == min(R)])
    
    beta_org <- as.matrix(BETA_F[-1, dd])
    
    beta0    <- BETA_F[1, dd]
    
    
  }
  
  
  temp     <- calbeta1_constraint_Subgradient(data, beta0, beta_org, alpha, KK, n_S, a_S, lambda_low, lambda_up, grid_number, S_lambda, approx_method)
  
  
  return( list(beta0 = beta0, beta1 = temp$beta1, lambda = temp$lambda) ) 
  
}

## IV.12, calculate another initial guess for DOSA, which removes subjects in tails before analysis.
## data: data set
## alpha: constraint for FPF
## KK: number of folds in cross-validation
## quan_num: the ratio of subjects in both tails to be removed before analysis

initial_remove_outlier <- function(data, alpha, KK, quan_num = 0.1){
  
  pp         <- dim(data$Marker1)[2]
  nn         <- dim(data$Marker1)[1]
  
  for (ii in 1:nn){
    
    
    for (jj in 1:pp){
      
      if (sum(is.na(data$Marker1[ii, ])) == 0){
        
        if(  (data$Marker1[ii, jj] > quantile(data$Marker1[ , jj], 1 - quan_num)) | (data$Marker1[ii, jj] < quantile(data$Marker1[ , jj], quan_num))    )
          
          data$Marker[ii, jj] <- NA
        
      }
      
      
    }
    
  }
  
  data  <- na.omit(data)
  
  beta  <- as.matrix(calpara_constraint_TPFP_mle(data, alpha, KK))
  
  return(beta)
  
}


#############################################################
#                                                           #
#            V. Standard Deviation                          #                            
#                                                           #
#############################################################

## V.1, computer variance of parameter beta and lambda.
## data: data set
## alpha: constraint for FPF
## beta_0: estimated intercept in the linear rule.
## beta_1: estimated slope in the linear rule.
## lambda: estimated lambda
## b: a pre-specified value for approximating the 0-1 loss by logit function
## approx_method: different surrogate functions


cal_variance_parameter_constraint <- function(data, alpha, beta0, beta1, lambda, b, approx_method = "hinge"){
  
  beta     <- rbind(as.matrix(beta0), beta1)
  data$f   <- data$Marker %*% beta
  pp       <- dim(data$Marker1)[2]
  
  data_1   <- data[data$Y==1, ]
  data_0   <- data[data$Y==0, ]
  
  n        <- dim(data)[1]
  n_1      <- dim(data_1)[1]
  n_0      <- dim(data_0)[1]
  
  data$f_index <- ifelse(data$f > 0, 1, 0)
  
  S_2       <- (1- data$Y) * as.matrix( data$f_index - alpha )/n_0
  
  V_S       <- matrix(rep(0, (pp+1)^2), pp+1, pp+1)
  
  H_11      <- matrix(rep(0, pp^2), pp, pp)
  
  H_21      <- t(as.matrix(rep(0, pp)))
  
  p         <- logistic(data$f/b)
  
  p_plus    <- logistic( (1 + data$f)/b)
  
  p_neg     <- logistic( (1 - data$f)/b)
  
  
  if (approx_method == "logistic") {
    
    for (ii in 1:n) {
      
      S_i <- as.matrix(1/n_1 * cal_first_deriv_loss_logistic( data$f[ii])*(data$Y[ii])* data$Marker1[ii, ]) -
        as.matrix(lambda/n_0 * cal_first_deriv_loss_logistic( - data$f[ii])*(1 - data$Y[ii])* data$Marker1[ii, ])
      
      
      S_i <- rbind(S_i, S_2[ii])
      
      
      V_S <- V_S + S_i %*% t(S_i)
      
    } 
    
    H_11 <-  t(data$Marker1) %*% (matrix( rep((cal_second_deriv_loss_logistic( data$f)*data$Y/n_1), pp), ncol = pp) * data$Marker1) +
      lambda* t(data$Marker1) %*% (matrix( rep((cal_second_deriv_loss_logistic(- data$f)*( 1 - data$Y)/n_0), pp), ncol = pp) * data$Marker1) 
    
    H_12 <-  - lambda/n_0 * t(data$Marker1) %*% (cal_first_deriv_loss_logistic( - data$f) * ( 1 - data$Y) )
    H_21 <-  t(data$Marker1) %*% (p*(1-p)*(1 - data$Y))/ b/n_0
    
    H  <- rbind(cbind(H_11, H_12), cbind(t(H_21), 0))
    
    
    
  } else if (approx_method == "hinge") {
    
    
    
    for (ii in 1:n) {
      
      S_i <- as.matrix(1/n_1 * cal_first_deriv_loss_hinge( data$f[ii])*(data$Y[ii])* data$Marker1[ii, ]) -
        as.matrix(lambda/n_0 * cal_first_deriv_loss_hinge(- data$f[ii])*(1 - data$Y[ii])* data$Marker1[ii, ])
      
      S_i <- rbind(S_i, S_2[ii])
      
      V_S <- V_S + S_i %*% t(S_i)
      
      
    }  
    
    
    H_11 <-  t(data$Marker1) %*% (matrix( rep((p_neg*(1 - p_neg)*data$Y/n_1/b), pp), ncol = pp) * data$ Marker1) +
      lambda * t(data$Marker1) %*% (matrix( rep((p_plus*(1 - p_plus)*( 1 - data$Y)/n_0/b), pp), ncol = pp) * data$ Marker1) 
    
    H_12 <-  - lambda/n_0 * t(data$Marker1) %*% (cal_first_deriv_loss_hinge( - data$f) * ( 1 - data$Y) )
    
    
    H_21 <-  t(data$Marker1) %*% (p*(1-p)*(1 - data$Y))/ b/n_0
    
    H  <- rbind(cbind(H_11, H_12), cbind(t(H_21), 0))
    
    
  }
  
  
  V <- solve(H) %*% V_S %*% t(solve(H))
  
  
  return( V = V )
  
}


## V.2, computer TPF and FPF for given BETA.
## data: data set
## BETA: a matrix contains a series of beta

calTPFP_linklinear    <- function(data, BETA) {
  
  pp           <- dim(BETA)[2]
  TP           <- as.matrix(rep(0, pp))
  FP           <- as.matrix(rep(0, pp))
  
  for (ii in 1:pp ) {
    
    beta         <- BETA[, ii]
    f            <- data$Marker %*% beta
    
    data$f_index <- ifelse(f > 0, 1, 0)
    
    data_1       <- data[data$Y==1, ]
    data_0       <- data[data$Y==0, ]
    
    TP[ii]       <- mean(data_1$f_index)
    FP[ii]       <- mean(data_0$f_index)
    
    
  }
  
  return( list(TP = TP, FP = FP) )
  
}


## V.3, computer first derivative of TP over beta1
## data: data set
## beta0: intercept in the linear rule.
## beta1: slope in the linear rule.
## delta_beta: distance of moving beta1 around

cal_first_deriv_TP_beta  <- function(data, beta0, beta1, delta_beta) {
  
  J          <- length(beta1)
  TP_deriv   <- as.matrix(rep(NA, J))
  
  for (jj in 1:J) {
    temp_delta         <- delta_beta
    
    beta1_temp_up      <- beta1
    beta1_temp_up[jj]  <- beta1[jj] + temp_delta
    
    beta1_temp_low     <- beta1
    beta1_temp_low[jj] <- beta1[jj] - temp_delta
    
    
    Delta_TP_j     <- caldiff_TPFP_linklinear(data, rbind(as.matrix(beta0), beta1_temp_up), 1)$TP - 
      caldiff_TPFP_linklinear(data, rbind(as.matrix(beta0), beta1_temp_low), 1)$TP 
    
    while (Delta_TP_j == 0 ) {
      temp_delta   <- temp_delta * 2
      
      beta1_temp_up      <- beta1
      beta1_temp_up[jj]  <- beta1[jj] + temp_delta
      
      beta1_temp_low     <- beta1
      beta1_temp_low[jj] <- beta1[jj] - temp_delta
      
      Delta_TP_j     <- caldiff_TPFP_linklinear(data, rbind(as.matrix(beta0), beta1_temp_up), 1)$TP - 
        caldiff_TPFP_linklinear(data, rbind(as.matrix(beta0), beta1_temp_low), 1)$TP 
      
      
    }
    
    
    TP_deriv[jj]       <- Delta_TP_j/2/delta_beta
    
  }
  
  
  return(TP_deriv = TP_deriv)
  
  
  
}

## V.4, computer the score and Hessian matrix of parameter beta and lambda.
## data: data set
## alpha: constraint for FPF
## beta_0: estimated intercept in the linear rule.
## beta_1: estimated slope in the linear rule.
## lambda: estimated lambda
## b: a pre-specified value for approximating the 0-1 loss by logit function
## approx_method: different surrogate functions


cal_SH_parameter_constraint <- function(data, alpha, beta0, beta1, lambda, b, approx_method = "hinge"){
  
  beta     <- rbind(as.matrix(beta0), beta1)
  data$f   <- data$Marker %*% beta
  pp       <- dim(data$Marker1)[2]
  
  data_1   <- data[data$Y==1, ]
  data_0   <- data[data$Y==0, ]
  
  n        <- dim(data)[1]
  n_1      <- dim(data_1)[1]
  n_0      <- dim(data_0)[1]
  
  data$f_index <- ifelse(data$f > 0, 1, 0)
  
  S_2       <- (1- data$Y) * as.matrix( data$f_index - alpha )/n_0
  
  H_11      <- matrix(rep(0, pp^2), pp, pp)
  
  H_21      <- t(as.matrix(rep(0, pp)))
  
  S         <- matrix(rep(NA, (pp+1)*n), ncol = n)
  
  p         <- logistic(data$f/b)
  
  p_plus    <- logistic( (1 + data$f)/b)
  
  p_neg     <- logistic( (1 - data$f)/b)
  
  
  if (approx_method == "logistic") {
    
    for (ii in 1:n) {
      
      S_i <- as.matrix(1/n_1 * cal_first_deriv_loss_logistic( data$f[ii])*(data$Y[ii])* data$Marker1[ii, ]) -
        as.matrix(lambda/n_0 * cal_first_deriv_loss_logistic( - data$f[ii])*(1 - data$Y[ii])* data$Marker1[ii, ])
      
      
      S_i <- rbind(S_i, S_2[ii])
      
      S[, ii] <- S_i    
      
    } 
    
    H_11 <-  t(data$Marker1) %*% (matrix( rep((cal_second_deriv_loss_logistic( data$f)*data$Y/n_1), pp), ncol = pp) * data$Marker1) +
      lambda* t(data$Marker1) %*% (matrix( rep((cal_second_deriv_loss_logistic(- data$f)*( 1 - data$Y)/n_0), pp), ncol = pp) * data$Marker1) 
    
    H_12 <-  - lambda/n_0 * t(data$Marker1) %*% (cal_first_deriv_loss_logistic( - data$f) * ( 1 - data$Y) )
    H_21 <-  t(data$Marker1) %*% (p*(1-p)*(1 - data$Y))/ b/n_0
    
    H  <- rbind(cbind(H_11, H_12), cbind(t(H_21), 0))
    
    
    
  } else if (approx_method == "hinge") {
    
    
    
    for (ii in 1:n) {
      
      S_i <- as.matrix(1/n_1 * cal_first_deriv_loss_hinge( data$f[ii])*(data$Y[ii])* data$Marker1[ii, ]) -
        as.matrix(lambda/n_0 * cal_first_deriv_loss_hinge(- data$f[ii])*(1 - data$Y[ii])* data$Marker1[ii, ])
      
      S_i <- rbind(S_i, S_2[ii])
      
      S[, ii] <- S_i 
      
    }  
    
    
    H_11 <-  t(data$Marker1) %*% (matrix( rep((p_neg*(1 - p_neg)*data$Y/n_1/b), pp), ncol = pp) * data$ Marker1) +
      lambda * t(data$Marker1) %*% (matrix( rep((p_plus*(1 - p_plus)*( 1 - data$Y)/n_0/b), pp), ncol = pp) * data$ Marker1) 
    
    H_12 <-  - lambda/n_0 * t(data$Marker1) %*% (cal_first_deriv_loss_hinge( - data$f) * ( 1 - data$Y) )
    
    
    H_21 <-  t(data$Marker1) %*% (p*(1-p)*(1 - data$Y))/ b/n_0
    
    H  <- rbind(cbind(H_11, H_12), cbind(t(H_21), 0))
    
    
  }
  
  
  
  return( list(S = S, H = H))
  
}


## V.6, computer variance of TFP.
## data: data set
## alpha: constraint for FPF
## beta_0: estimated intercept in the linear rule.
## beta_1: estimated slope in the linear rule.
## lambda: estimated lambda
## b: a pre-specified value for approximating the 0-1 loss by logit function
## delta_beta: distance of moving beta1 around
## approx_method: different surrogate functions

cal_meanandsd_TP_constraint <- function( data, alpha, beta0, beta1, lambda, b, delta_beta, approx_method = "hinge" ) {
  
  
  n       <- dim(data)[1]
  n1      <- sum(data$Y)
  n0      <- n - n1
  pp      <- length(beta1)
  
  data$f  <- ifelse(data$Marker %*% rbind(as.matrix(beta0), beta1) > 0, 1, 0 )
  
  TP_m    <- sum(data$f * data$Y/n1)
  
  W1      <- (data$f/n1 - TP_m/n1)*data$Y
  
  temp    <- cal_SH_parameter_constraint(data, alpha, beta0, beta1, lambda, b, approx_method)
  
  T       <- (solve(temp$H))[1:pp, ]
  
  TP_deriv <- cal_first_deriv_TP_beta(data, beta0, beta1, delta_beta)
  
  W2      <- t(TP_deriv) %*% T %*% temp$S
  
  # W       <- W1 + t(W2)
  
  W       <- W1 - t(W2)
  
  var_TP  <- t(W) %*% W
  
  return(list(sd_TP = sqrt(var_TP), mean_TP = TP_m) )
  
}




